/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Abstract Cecilia compiler Launcher.
 */
public abstract class AbstractLauncher {

  private final CmdOption        srcPathOpt   = new CmdPathOption(
                                                  new String[]{"src-path", "S"},
                                                  "the search path of ADL,IDL and implementation files",
                                                  ".");

  private final CmdOption        outDirOpt    = new CmdOption(
                                                  new String[]{"out-path",
      "out-dir", "o"                              },
                                                  "the path where generated files will be put",
                                                  ".");

  private final CmdOption[]      coreOptions  = new CmdOption[]{srcPathOpt,
      outDirOpt                               };

  private static final String    IDL_DIR_NAME = "idl";
  private static final String    ADL_DIR_NAME = "adl";
  private static final String    OBJ_DIR_NAME = "obj";

  protected CmdFlag[]            allCmbOptions;

  // Map<String, CmdOption> of options
  protected Map<String, CmdFlag> options      = new HashMap<String, CmdFlag>();

  protected List<String>         adlNames     = new ArrayList<String>();

  // List<String> of source files and directory
  protected List<String>         sourcePaths;
  // source class loader
  protected ClassLoader          sourceClassLoader;

  // build directory
  protected File                 buildDir;
  // idl build directory
  protected File                 idlBuildDir;
  // adl build directory
  protected File                 adlBuildDir;
  // obj build directory
  protected File                 objBuildDir;

  protected void setSpecificOptions(final CmdFlag... specificOptions) {
    addOptions(coreOptions);
    addOptions(specificOptions);
    allCmbOptions = new CmdFlag[coreOptions.length + specificOptions.length];
    System.arraycopy(coreOptions, 0, allCmbOptions, 0, coreOptions.length);
    System.arraycopy(specificOptions, 0, allCmbOptions, coreOptions.length,
        specificOptions.length);
  }

  protected abstract void printUsage(PrintStream ps);

  protected List<String> parsePathList(String paths) {
    final List<String> l = new ArrayList<String>();
    int index = paths.indexOf(File.pathSeparatorChar);
    while (index != -1) {
      l.add(paths.substring(0, index));
      paths = paths.substring(index + 1);
      index = paths.indexOf(File.pathSeparatorChar);
    }
    l.add(paths);
    return l;
  }

  protected void parseArgs(final String... args)
      throws InvalidCommandLineException {
    for (final String arg : args) {
      if (arg.startsWith("-")) {
        final int index = arg.indexOf('=');
        if (index == -1) {
          final CmdFlag cmdFlag = options.get(arg.substring(1));
          if (cmdFlag instanceof CmdOption)
            throw new InvalidCommandLineException("Invalid option: '" + arg
                + "' must be followed by a value", 1);
          else
            cmdFlag.setPresent();
        } else {
          if (index < 2) {
            throw new InvalidCommandLineException("Invalid option '" + arg
                + "'", 1);
          }
          final String optionName = arg.substring(1, index);
          final String optionValue = arg.substring(index + 1);

          // validate option
          final CmdFlag cmdFlag = options.get(optionName);
          if (cmdFlag == null) {
            throw new InvalidCommandLineException("Unknown option '" + arg
                + "'", 1);
          }
          if (!(cmdFlag instanceof CmdOption))
            throw new InvalidCommandLineException("Invalid option: '" + arg
                + "' can't be followed by a value", 1);
          else
            ((CmdOption) cmdFlag).setValue(optionValue);
        }
      } else {
        adlNames.add(arg);
      }
    }
    prepareOutDir();
    prepareSourceClassLoader();
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  private void addOptions(final CmdFlag[] options) {
    for (final CmdFlag option : options) {
      for (final String optName : option.options) {
        if (this.options.put(optName, option) != null) {
          throw new Error("Duplicated option '" + optName + "'");
        }
      }
    }
  }

  private void prepareSourceClassLoader() {
    final String srcPath = srcPathOpt.getValue();
    final List<String> paths = parsePathList(srcPath);
    sourcePaths = new ArrayList<String>(paths.size());

    // check source paths
    for (final String path : paths) {
      final File f = new File(path);
      if (!f.exists()) {
        System.out.println("Warning '" + f.getAbsolutePath()
            + "' source path can't be found ");
      } else if (!f.isDirectory()) {
        System.out.println("Warning: \"" + path
            + "\" is not a directory, path ignored.");
      } else {
        sourcePaths.add(path);
      }
    }

    // add "build/adl" and "build/idl" as source path
    sourcePaths.add(idlBuildDir.getAbsolutePath());
    sourcePaths.add(adlBuildDir.getAbsolutePath());

    // build URL array of source path
    final URL[] urls = new URL[sourcePaths.size()];
    for (int i = 0; i < urls.length; i++) {
      final String path = sourcePaths.get(i);
      final File f = new File(path);
      try {
        urls[i] = f.toURL();
      } catch (final MalformedURLException e) {
        // never append
        throw new RuntimeException(e);
      }
    }

    sourceClassLoader = new URLClassLoader(urls, getClass().getClassLoader());
  }

  private void prepareOutDir() throws InvalidCommandLineException {
    String outDirName = outDirOpt.getValue();
    outDirName = outDirName.trim();
    buildDir = new File(outDirName);
    if (!buildDir.exists())
      throw new InvalidCommandLineException("Invalid out directory '"
          + buildDir.getAbsolutePath() + "'. Directory not found", 2);
    if (!buildDir.isDirectory())
      throw new InvalidCommandLineException("Invalid out directory '"
          + buildDir.getAbsolutePath() + "'. Not a directory", 2);

    idlBuildDir = checkBuildDir(buildDir, IDL_DIR_NAME);
    adlBuildDir = checkBuildDir(buildDir, ADL_DIR_NAME);
    objBuildDir = checkBuildDir(buildDir, OBJ_DIR_NAME);
  }

  private File checkBuildDir(final File outDir, final String dirName)
      throws InvalidCommandLineException {
    final File d = new File(outDir, dirName);
    if (d.exists()) {
      if (!d.isDirectory())
        throw new InvalidCommandLineException(
            "Invalid out directory '" + outDir.getAbsolutePath() + "'. '"
                + dirName + "' not a directory", 6);
    } else {
      if (!d.mkdir())
        throw new InvalidCommandLineException("Invalid out directory '"
            + outDir.getAbsolutePath() + "'. Unable to create'" + dirName
            + "' directory", 7);
    }
    return d;
  }

  protected void printOptions(final PrintStream ps) {
    printUsage(ps);
    ps.println();
    ps.println("Available options are :");
    int maxCol = 0;

    for (final CmdFlag opt : allCmbOptions) {
      int col = 3 + opt.options[0].length();
      if (opt instanceof CmdOption) col += 4;
      if (opt.options.length > 1) {
        col += opt.options[1].length() + 2;
        if (opt.options.length > 2)
          for (int j = 2; j < opt.options.length; j++) {
            col += opt.options[j].length() + 2;
          }
        else
          col += 2;
      }
      if (col > maxCol) maxCol = col;
    }
    for (final CmdFlag opt : allCmbOptions) {
      final StringBuffer sb = new StringBuffer("  -");
      sb.append(opt.options[0]);
      if (opt instanceof CmdOption) {
        sb.append("=ARG");
      }
      if (opt.options.length > 1) {
        sb.append(" (");
        for (int j = 1; j < opt.options.length; j++) {
          sb.append(opt.options[j]);
          if (j < opt.options.length - 1) sb.append(", ");
        }
        sb.append(") ");
      }
      while (sb.length() < maxCol)
        sb.append(' ');
      sb.append(": ").append(opt.getInfo());
      ps.println(sb);
    }
  }

  // ---------------------------------------------------------------------------
  // Internal classes
  // ---------------------------------------------------------------------------

  /**
   * Exception thrown when an error on the command line has been detected.
   */
  public static class InvalidCommandLineException extends Exception {

    final int exitValue;

    /**
     * @param message detail message.
     * @param exitValue exit value.
     */
    public InvalidCommandLineException(final String message, final int exitValue) {
      super(message);
      this.exitValue = exitValue;
    }

  }

  protected static class CmdFlag {
    protected final String[] options;
    protected final String   info;
    protected boolean        present;

    protected CmdFlag(final String[] options, final String info) {
      this.options = options;
      this.info = info;
    }

    protected void setPresent() {
      present = true;
    }

    protected boolean isPresent() {
      return present;
    }

    /**
     * @return options
     */
    public String[] getOptions() {
      return options;
    }

    /**
     * @return detail message
     */
    public String getInfo() {
      return info;
    }
  }

  protected static class CmdOption extends CmdFlag {
    protected String value;

    protected CmdOption(final String[] options, final String info,
        final String defaultValue) {
      super(options, info);
      this.value = defaultValue;
    }

    protected CmdOption(final String[] options, final String info) {
      this(options, info, null);
    }

    protected CmdOption(final String option, final String info) {
      this(new String[]{option}, info, null);
    }

    protected final String getValue() {
      return value;
    }

    protected void setValue(final String value) {
      setPresent();
      this.value = value;
    }
  }

  protected static class CmdAppendOption extends CmdOption {

    protected CmdAppendOption(final String option, final String info) {
      super(option, info);
    }

    protected CmdAppendOption(final String[] options, final String info,
        final String initialValue) {
      super(options, info, initialValue);
    }

    protected CmdAppendOption(final String[] options, final String info) {
      super(options, info);
    }

    /** Append new value at end. */
    @Override
    protected void setValue(final String value) {
      setPresent();
      if (value == null) return;
      this.value = (this.value == null) ? value : this.value + " " + value;
    }

    @Override
    public String getInfo() {
      return super.getInfo() + "; this option may be specified several time";
    }
  }

  protected static class CmdPathOption extends CmdOption {

    protected CmdPathOption(final String option, final String info) {
      super(option, info);
    }

    protected CmdPathOption(final String[] options, final String info,
        final String initialValue) {
      super(options, info, initialValue);
    }

    protected CmdPathOption(final String[] options, final String info) {
      super(options, info);
    }

    /** Append new value at end. */
    @Override
    protected void setValue(String value) {
      setPresent();
      if (value == null) return;
      value = value.trim();
      this.value = (this.value == null) ? value : this.value
          + File.pathSeparator + value;
    }

    @Override
    public String getInfo() {
      return super.getInfo() + "; this option may be specified several time";
    }
  }
}
