/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.PrintStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FactoryFactory;
import org.objectweb.fractal.adl.JavaFactory;
import org.objectweb.fractal.adl.StaticJavaGenerator;
import org.objectweb.fractal.adl.util.FractalADLLogManager;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.plugin.AbstractPluginManager;
import org.objectweb.fractal.util.Fractal;

/**
 * A class to compile a component. Usage: Launcher &lt;definition&gt; where
 * &lt;definition&gt; is the name of the component to be compiled.
 */
public class Launcher extends AbstractLauncher {

  /**
   * The name of the {@link System#getProperty(String) system property} that
   * specify the backend to use for the plugin factory.
   */
  public static final String PLUGIN_FACTORY_BACKEND_PROPERTY_NAME = AbstractPluginManager.PLUGIN_FACTORY_BACKEND;

  /**
   * The default value of the {@link #PLUGIN_FACTORY_BACKEND_PROPERTY_NAME}
   * property.
   */
  public static final String DEFAULT_PLUGIN_FACTORY_BACKEND       = "Java";

  /** The name of the default ADL of the compiler. */
  public static final String CECILIA_BASIC_COMPILER_ADL           = "org.objectweb.fractal.cecilia.adl.CeciliaBasicFactory";

  protected final CmdOption  compilerADLOpt                       = new CmdOption(
                                                                      new String[]{"compiler-adl"},
                                                                      "Specify the name of the ADL of the compiler itself",
                                                                      CECILIA_BASIC_COMPILER_ADL);

  protected final CmdOption  compilerCmdOpt                       = new CmdOption(
                                                                      new String[]{
      "compiler-command", "cc"                                        },
                                                                      "the command of the C compiler (default is 'gcc')",
                                                                      "gcc");

  protected final CmdOption  defsOpt                              = new CmdAppendOption(
                                                                      new String[]{
      "defs", "D"                                                     },
                                                                      "the -D compiler directives");
  protected final CmdOption  cFlagsOpt                            = new CmdAppendOption(
                                                                      new String[]{
      "c-flags", "cf"                                                 },
                                                                      "the c-flags compiler directives");

  protected final CmdOption  includePathOpt                       = new CmdPathOption(
                                                                      new String[]{
      "inc-path", "I"                                                 },
                                                                      "the -I compiler directives");

  protected final CmdOption  linkerCmdOpt                         = new CmdOption(
                                                                      new String[]{
      "linker-command", "ld"                                          },
                                                                      "the command of the linker (default is 'gcc')",
                                                                      "gcc");

  protected final CmdOption  executableFormat                     = new CmdOption(
                                                                      new String[]{"executable-format"},
                                                                      "the binary format associated to \"executable\" output format (default is 'elf')",
                                                                      "elf");

  protected final CmdOption  ldFlagsOpt                           = new CmdAppendOption(
                                                                      new String[]{
      "ld-flags", "lf"                                                },
                                                                      "the ld-flags compiler directives");
  protected final CmdOption  ldPathOpt                            = new CmdPathOption(
                                                                      new String[]{
      "ld-path", "-L"                                                 },
                                                                      "the -L compiler directives");

  protected final CmdOption  archiveCmdOpt                        = new CmdOption(
                                                                      new String[]{
      "archive-command", "ar"                                         },
                                                                      "the command of the archive tool, (default is 'ar')",
                                                                      "ar");

  protected final CmdOption  linkerScriptOpt                      = new CmdOption(
                                                                      new String[]{
      "linker-script", "T"                                            },
                                                                      "linker script");

  protected final CmdOption  executableNameCmdOpt                 = new CmdOption(
                                                                      new String[]{
      "executable-name", "exec-name"                                  },
                                                                      "the name of the target executable file (can be used with only one ADL to be compiled.)");

  protected final CmdOption  concurrentJobCmdOpt                  = new CmdOption(
                                                                      new String[]{
      "jobs", "j"                                                     },
                                                                      "The number of concurrent compilation jobs",
                                                                      "1");

  protected final CmdFlag    printStackTraces                     = new CmdFlag(
                                                                      new String[]{"e"},
                                                                      "Print error stack traces");

  Map<String, String>        adlToExecName                        = new LinkedHashMap<String, String>();
  Map<String, Object>        compilerContext                      = new HashMap<String, Object>();
  Factory                    compiler;

  private static Logger      logger                               = FractalADLLogManager
                                                                      .getLogger("launcher");

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param outDir the output directory
   * @param sources the source path.
   * @throws InvalidCommandLineException
   * @throws CompilerInstantiationException
   */
  public Launcher(final String outDir, final String sources)
      throws InvalidCommandLineException, CompilerInstantiationException {
    init("-o=" + outDir, "-src-path=" + sources);
  }

  /**
   * Creates a new Compiler launcher context.
   * 
   * @param options the options
   * @throws InvalidCommandLineException
   * @throws CompilerInstantiationException
   */
  public Launcher(final Map<String, String> options)
      throws InvalidCommandLineException, CompilerInstantiationException {
    init(toArgs(options));
  }

  protected static String[] toArgs(final Map<String, String> options) {
    final String[] params = new String[options.size()];
    int i = 0;
    for (final Map.Entry<String, String> option : options.entrySet()) {
      params[i] = option.getKey() + '=' + option.getValue();
      i++;
    }
    return params;
  }

  /**
   * Creates a new Compiler launcher context and run compilation.
   * 
   * @param args the command line arguments.
   * @throws Exception
   */
  public Launcher(final String... args) throws Exception {
    try {
      init(args);
      compile();
    } catch (final InvalidCommandLineException e) {
      handleException(e);
    } catch (final CompilerInstantiationException e) {
      handleException(e);
    } catch (final ADLException e) {
      handleException(e);
    }
  }

  protected void init(final String... args) throws InvalidCommandLineException,
      CompilerInstantiationException {
    if (logger.isLoggable(Level.CONFIG)) {
      for (final String arg : args) {
        logger.config("[arg] " + arg);
      }
    }

    setSpecificOptions(compilerADLOpt, compilerCmdOpt, defsOpt, cFlagsOpt,
        includePathOpt, linkerCmdOpt, ldFlagsOpt, ldPathOpt, executableFormat,
        archiveCmdOpt, linkerScriptOpt, executableNameCmdOpt,
        concurrentJobCmdOpt, printStackTraces);

    parseArgs(args);

    /*
     * get the CeciliaADL Compiler (default implementation a pure Java
     * component)
     */

    this.compiler = createCompiler(compilerContext);

    final List<String> incPaths = new ArrayList<String>(sourcePaths);
    if (this.includePathOpt.getValue() != null)
      incPaths.addAll(parsePathList(includePathOpt.getValue()));
    StringBuffer sb = new StringBuffer();
    for (final String inc : incPaths) {
      final File incDir = new File(inc);
      sb.append(" -I").append(incDir.getAbsolutePath());
    }
    final String includePath = sb.toString();

    final List<String> ldPaths = new ArrayList<String>();
    if (ldPathOpt.getValue() != null)
      ldPaths.addAll(parsePathList(ldPathOpt.getValue()));
    sb = new StringBuffer();
    for (final String ld : ldPaths) {
      final File ldDir = new File(ld);
      sb.append(" -L").append(ldDir.getAbsolutePath());
    }
    final String ldPath = sb.toString();

    // add compiler arguments to context
    String optValue = compilerCmdOpt.getValue();
    if (nullOrEmpty(optValue)) {
      throw new InvalidCommandLineException("Invalid compiler ''", 1);
    }
    compilerContext.put(compilerCmdOpt.options[0], optValue);

    optValue = cFlagsOpt.getValue();
    if (!nullOrEmpty(optValue)) {
      compilerContext.put(cFlagsOpt.options[0], optValue);
    }

    compilerContext.put(includePathOpt.options[0], includePath);
    optValue = defsOpt.getValue();
    if (!nullOrEmpty(optValue)) {
      compilerContext.put(defsOpt.options[0], optValue);
    }

    optValue = linkerCmdOpt.getValue();
    if (nullOrEmpty(optValue)) {
      throw new InvalidCommandLineException("Invalid linker ''", 1);
    }
    compilerContext.put(linkerCmdOpt.options[0], optValue);

    optValue = ldFlagsOpt.getValue();
    if (!nullOrEmpty(optValue)) {
      compilerContext.put(ldFlagsOpt.options[0], optValue);
    }

    compilerContext.put(ldPathOpt.options[0], ldPath);
    compilerContext.put(archiveCmdOpt.options[0], archiveCmdOpt.getValue());
    compilerContext.put(executableFormat.options[0], executableFormat
        .getValue());

    Integer jobs = null;
    try {
      jobs = Integer.decode(concurrentJobCmdOpt.getValue());
    } catch (final NumberFormatException e) {
      throw new InvalidCommandLineException("Invalid jobs value '"
          + concurrentJobCmdOpt.getValue() + "' is not a valid number", 1);
    }
    compilerContext.put(concurrentJobCmdOpt.options[0], jobs);

    // add build directories to context
    compilerContext.put("adlBuildDirectory", adlBuildDir);
    compilerContext.put("idlBuildDirectory", idlBuildDir);
    compilerContext.put("objBuildDirectory", objBuildDir);

    // add source class loader in context
    compilerContext.put("classloader", sourceClassLoader);

    // add linker script to the context
    final String linkerScript = linkerScriptOpt.getValue();
    if (linkerScript != null) {
      final URL linkerScriptURL = sourceClassLoader.getResource(linkerScript);
      if (linkerScriptURL == null) {
        throw new InvalidCommandLineException("Invalid linker script: '"
            + linkerScript + "'. Cannot find file in the source path", 1);
      }

      compilerContext
          .put(linkerScriptOpt.options[0], linkerScriptURL.getPath());
    }
  }

  @Override
  protected void parseArgs(final String... args)
      throws InvalidCommandLineException {
    super.parseArgs(args);

    boolean defaultExeNameUsed = false;
    // parse adlNames
    for (final String adlName : adlNames) {
      final int i = adlName.indexOf(':');
      if (i == -1) {
        if (defaultExeNameUsed) {
          if (executableNameCmdOpt.getValue() == null)
            throw new InvalidCommandLineException(
                "Executable name must be specifyed for adl \"" + adlName
                    + "\" using -" + executableNameCmdOpt.options[0]
                    + "=NAME or the ADL:NAME syntax.", 1);
          else
            throw new InvalidCommandLineException(
                "Executable name must be specifyed for adl \"" + adlName
                    + "\" using the ADL:NAME syntax.", 1);
        }
        adlToExecName.put(adlName, executableNameCmdOpt.getValue());
        defaultExeNameUsed = true;
      } else {
        final String adl = adlName.substring(0, i);
        final String exec = adlName.substring(i + 1);
        adlToExecName.put(adl, exec);
      }
    }
  }

  /**
   * Create the o.o.f.a.Factory class that will be used to read the application
   * definition. Subclasses of the Launcher may extend this.
   * 
   * @param compilerDefinition
   * @param compilerContext
   * @return
   * @throws CompilerInstantiationException
   * @throws ADLException
   * @throws Exception
   * @throws Exception
   */
  protected Factory createCompiler(final Map<String, Object> compilerContext)
      throws CompilerInstantiationException {
    final String compilerAdlDefinition = compilerADLOpt.getValue();

    final Object pluginFactoryBackend = compilerContext
        .get(PLUGIN_FACTORY_BACKEND_PROPERTY_NAME);
    if (pluginFactoryBackend == null
        || DEFAULT_PLUGIN_FACTORY_BACKEND.equals(pluginFactoryBackend)) {
      /* instantiate the bootstrap factory with the default backend */
      try {
        final Class<?> factoryFactoryClass = getClass().getClassLoader()
            .loadClass(
                compilerAdlDefinition + StaticJavaGenerator.CLASSNAME_SUFFIX);
        final JavaFactory factoryFactory = (JavaFactory) factoryFactoryClass
            .newInstance();
        final Map<?, ?> compilerComponent = (Map<?, ?>) factoryFactory
            .newComponent();
        logger
            .fine("[Launcher] Cecilia compiler instantiated using statically generated factory.");
        return (Factory) compilerComponent.get("factory");
      } catch (final Exception e) {
        // ignore and try to instantiate the compiler using a FractalADL
        // factory.
        if (logger.isLoggable(Level.FINE)) {
          logger
              .fine("[Launcher] Unable to instantiate cecilia compiler using statically generated factory ("
                  + e.getClass().getName() + ":" + e.getMessage() + ").");
        }
      }

      try {
        final Factory factory = FactoryFactory.getFactory();
        /* try to instantiate the CeciliaADL BasicFactory */
        final Map<?, ?> compilerComponent = (Map<?, ?>) factory.newComponent(
            compilerAdlDefinition, compilerContext);
        logger
            .fine("[Launcher] Cecilia compiler instantiated using FractalADL factory.");
        return (Factory) compilerComponent.get("factory");
      } catch (final ADLException e) {
        throw new CompilerInstantiationException(
            "Unable to instantiate Cecilia compiler", e, 101);
      }

    } else {
      /* instantiate the bootstrap factory with the Fractal backend */
      Component compilerComponent = null;
      try {
        final Factory factory = FactoryFactory
            .getFactory(FactoryFactory.FRACTAL_BACKEND);

        /* try to instantiate the CeciliaADL BasicFactory */
        compilerComponent = (Component) factory.newComponent(
            compilerAdlDefinition, compilerContext);
        /* start compiler */
        Fractal.getLifeCycleController(compilerComponent).startFc();

        final Factory result = (Factory) compilerComponent
            .getFcInterface("factory");
        return result;
      } catch (final ADLException e) {
        throw new CompilerInstantiationException(
            "Unable to instantiate Cecilia compiler", e, 101);
      } catch (final NoSuchInterfaceException e) {
        throw new CompilerInstantiationException(
            "Unable to instantiate Cecilia compiler", e, 101);
      } catch (final IllegalLifeCycleException e) {
        throw new CompilerInstantiationException(
            "Unable to instantiate Cecilia compiler", e, 101);
      }
    }
  }

  private static boolean nullOrEmpty(final String s) {
    return s == null || s.length() == 0;
  }

  /**
   * Compiles the given ADL.
   * 
   * @param adlName the name of the the name of the architecture (ADL) to
   *            compile.
   * @return the list of objects that are the results of the compilation.
   * @throws ADLException if the compilation fails.
   */
  public List<Object> compile(final String adlName) throws ADLException {
    final List<Object> result = new ArrayList<Object>();
    compile(adlName, executableNameCmdOpt.getValue(), result);
    return result;
  }

  /**
   * Compiles the given ADL.
   * 
   * @param adlName the name of the architecture (ADL) to compile.
   * @param execName the name of the executable file to be created.
   * @return the list of objects that are the results of the compilation.
   * @throws ADLException if the compilation fails.
   */
  public List<Object> compile(final String adlName, final String execName)
      throws ADLException {
    final List<Object> result = new ArrayList<Object>();
    compile(adlName, execName, result);
    return result;
  }

  /**
   * Compiles the list of adl files that are received from the command-line.
   * 
   * @return the list of objects that are the results of the compilation of the
   *         list of adl files received from the command-line.
   * @throws ADLException if the compilation fails.
   * @throws InvalidCommandLineException
   */
  public List<Object> compile() throws ADLException,
      InvalidCommandLineException {
    // Check if at least 1 adlName is specified
    if (adlToExecName.size() == 0) {
      throw new InvalidCommandLineException("no definition name is specified.",
          1);
    }

    final List<Object> result = new ArrayList<Object>();
    for (final Map.Entry<String, String> e : adlToExecName.entrySet()) {
      compile(e.getKey(), e.getValue(), result);
    }
    return result;
  }

  private void compile(final String adlName, final String execName,
      final List<Object> result) throws ADLException {
    final HashMap<Object, Object> contextMap = new HashMap<Object, Object>(
        compilerContext);
    if (execName != null)
      contextMap.put(executableNameCmdOpt.options[0], execName);
    else
      contextMap
          .put(executableNameCmdOpt.options[0], adlName.replace('.', '_'));

    final Object compilationResult = this.compiler.newComponent(adlName,
        contextMap);

    if (compilationResult instanceof File) {
      result.add(compilationResult);
    } else if (compilationResult instanceof Collection) {
      for (final Object elem : (Collection<?>) compilationResult) {
        result.add(elem);
      }
    }
  }

  @Override
  protected void printUsage(final PrintStream ps) {
    ps.println("Usage: Launcher [OPTIONS] (<definition>[:<execname>])+");
    ps.println("  where <definition> is the name of the component to"
        + " be compiled, ");
    ps
        .println("  and <execname> is the name of the output file to be created.");
  }

  protected void handleException(final InvalidCommandLineException e)
      throws Exception {
    System.err.println(e.getMessage());
    printOptions(System.err);
    System.exit(e.exitValue);
  }

  protected void handleException(final CompilerInstantiationException e)
      throws Exception {
    e.printStackTrace();
    System.exit(e.exitValue);
  }

  protected void handleException(final ADLException e) throws Exception {
    if (printStackTraces.isPresent()) {
      e.printStackTrace();
    } else {
      final StringBuffer sb = new StringBuffer();
      sb.append(e.getMessage()).append('\n');
      Throwable cause = e.getCause();
      while (cause != null) {
        sb.append("caused by : ");
        sb.append(cause.getMessage()).append('\n');
        cause = cause.getCause();
      }
      System.err.println(sb);
    }
    System.exit(1);
  }

  /**
   * Entry point.
   * 
   * @param args
   */
  public static void main(final String... args) {
    try {
      new Launcher(args);
    } catch (final Exception e) {
      // never append
      e.printStackTrace();
    }
  }

  /**
   * Exception thrown when the CeciliaADL compiler can't be instantiated.
   */
  public static class CompilerInstantiationException extends Exception {

    final int exitValue;

    /**
     * @param message detail message.
     * @param cause cause.
     * @param exitValue exit value.
     */
    public CompilerInstantiationException(final String message,
        final Throwable cause, final int exitValue) {
      super(message, cause);
      this.exitValue = exitValue;
    }

  }
}
